from typing import List, Dict, Callable, Any
from cdh_nav_scraper import CDHNavgatorScraper
from data_util import Data, DataUsage
import ray
import json
import export_util
import hive_util
import log_util
import option_util

ITEM_COUNT_PER_PAGE = 25
logger = log_util.get_default_logger()
pattern = "(?P<data_type>[a-z]+).(?P<service>[a-z]+)_(?P<module>[a-z0-9_]+)_(?P<version>v[0-9]+)"


def parse_args() -> Dict[str, Any]:
    parser = option_util.get_parser_with_default_args()
    parser.add_argument("--search-facet-started", type=str, default="[NOW/DAY-1DAYS TO NOW/DAY+1DAY]",
                        help="started date to search",)
    parser.add_argument("--search-facet-deleted", type=str, default="false",
                        help="including deleted data in search result")
    parser.add_argument("--search-facet-source-type", type=str, default="hive", help="source type to search")
    parser.add_argument("--search-facet-type", type=str, default="operation_execution", help="data type to search")
    parser.add_argument("--extra-search-facet", type=str, help="extra search facets")
    parser.add_argument("--cdh-username", type=str, help="CDH username", required=True)
    parser.add_argument("--cdh-password", type=str, help="CDH password", required=True)
    parser.add_argument("--cdh-nav-url", type=str, default="http://ahdp138-glace.nfra.io:7187",
                        help="CDH navigator URL")
    parser.add_argument("--chrome-driver-path", type=str, help="Chrome driver file path", default="./chromedriver")
    parser.add_argument("--selenium-timeout", type=int, help="timeout for selenium(sec)", default=300)
    return vars(parser.parse_args())


def generate_search_facets(search_facet_started: str, search_facet_deleted: str, search_facet_source_type: str,
                           search_facet_type: str, extra_search_facets: str) -> str:
    result = {}
    facets = [
        ("started", search_facet_started),
        ("deleted", search_facet_deleted),
        ("sourceType", search_facet_source_type),
        ("type", search_facet_type)
    ]
    if extra_search_facets:
        for extra in extra_search_facets.split(" "):
            key_and_value = extra.split(":")
            if len(key_and_value) != 2:
                raise "invalid EXTRA_SEARCH_FACETS"
            facets.append(tuple(key_and_value))
    for facet in facets:
        result[facet[0]] = facet[1].split(",")
    return json.dumps(result, separators=(',', ':'))


def get_hive_tables(hive_uri: str, hive_username: str, hive_password: str, database: str) -> List[str]:
    query_result = hive_util.query_to_hive(hive_uri, hive_username, hive_password, "show tables", database)
    tables = query_result.get("tab_name")
    return tables.values.tolist()


def scrape(scrape_func: Callable, chrome_driver_path: str, selenium_timeout: int,
           cdh_nav_url: str, cdh_username, cdh_password):
    scraper = CDHNavgatorScraper(chrome_driver_path, selenium_timeout, logger)
    try:
        scraper.login(f"{cdh_nav_url}/login.html", cdh_username, cdh_password)
        return scrape_func(scraper)
    finally:
        scraper.close()


@ray.remote
def scrape_search_results_async(thread: int, num_threads: int, cdh_nav_url: str, cdh_username: str, cdh_password: str,
                                chrome_driver_path: str, selenium_timeout: int, search_facets: str) -> Dict[str, int]:
    def func(scraper: CDHNavgatorScraper):
        search_result_count = scraper.get_search_result_count(f"{cdh_nav_url}/?facets={search_facets}")
        operation_execution_urls = []
        offset = thread * ITEM_COUNT_PER_PAGE
        while offset < search_result_count:
            logger.info(f"scraping search results... [offset: {offset}, total:{search_result_count}, thread: {thread}]")
            operation_execution_urls.extend(
                scraper.scrape_search_results(f"{cdh_nav_url}/?facets={search_facets}&offset={offset}")
            )
            offset += num_threads * ITEM_COUNT_PER_PAGE
        return scraper.scrape_operation_executions(operation_execution_urls)
    return scrape(func, chrome_driver_path, selenium_timeout, cdh_nav_url, cdh_username, cdh_password)


def main():
    args = parse_args()
    num_threads = args.get("num_threads")
    cdh_nav_url = args.get("cdh_nav_url")
    cdh_username = args.get("cdh_username")
    cdh_password = args.get("cdh_password")
    chrome_driver_path = args.get("chrome_driver_path")
    selenium_timeout = args.get("selenium_timeout")

    search_facets = generate_search_facets(args.get("search_facet_started"), args.get("search_facet_deleted"),
                                           args.get("search_facet_source_type"), args.get("search_facet_type"),
                                           args.get("extra_search_facets"))

    ray_ids = [scrape_search_results_async.remote(i, num_threads, cdh_nav_url, cdh_username,
                                                  cdh_password, chrome_driver_path,
                                                  selenium_timeout, search_facets) for i in range(0, num_threads)]
    scrape_results = []
    while len(ray_ids):
        done, ray_ids = ray.wait(ray_ids)
        for scrape_result in ray.get(done):
            scrape_results.append(scrape_result)

    url_and_usage = {}
    for scrape_result in scrape_results:
        for key in scrape_result.keys():
            if key not in url_and_usage.keys():
                url_and_usage[key] = scrape_result[key]
            else:
                url_and_usage[key] += scrape_result[key]
    databases = ["snapshot", "compact"]

    def scrape_table_info(scraper: CDHNavgatorScraper) -> DataUsage:
        data_usage = DataUsage()
        for database in databases:
            query_results = get_hive_tables(args.get("hive_uri"), args.get("hive_username"),
                                            args.get("hive_password"), database)
            for table in query_results:
                data = Data.from_pattern(pattern, f"{database}.{table}")
                data_usage.set_usage_with_data(0, data)
        return scraper.scrape_table_info(url_and_usage, data_usage, pattern)

    result = scrape(scrape_table_info, chrome_driver_path, selenium_timeout,
                    cdh_nav_url, cdh_username, cdh_password)
    columns = ["total"]
    columns.extend(databases)
    export_util.export_to_table(result, columns, ["total"], "cdh-nav")
    export_util.export_unused_items_to_checklist(result, "cdh-nav")


if __name__ == "__main__":
    main()
