from sql_metadata import Parser
from typing import List, Dict
import re

setting_pattern = "(set|SET)[ ]+(?P<key>[a-z.]+)=(?P<value>[a-zA-z0-9,_-]+)"
session_id_pattern = "(?P<session_id>[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{8})"
principal_pattern = "ugi=(?P<principal>[a-z.-_]+)"
timestamp_pattern = "^(?P<timestamp>[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2},[0-9]{3})"
timestamp_date_format = "2022-06-13T01:31:25,562"
use_database_query_pattern = "(use|USE)\\s+`?(?P<database>[a-z0-9_]+)`?"
query_pattern = f"Executing command\\(queryId=[a-z0-9-_]+\\): (?P<query>[\\w\\W\\s]+)"
db_and_table_pattern = "[a-z0-9_-]+.[a-z0-9_-]+"


class QueryHistory:

    def __init__(self, raw_hiveql: str, settings: Dict[str, str], is_supported: bool,
                 tables: List[str], principal: str, timestamp: str, database: str):
        self.raw_hiveql = raw_hiveql
        self.settings = settings
        self.tables = tables
        self.principal = principal
        self.timestamp = timestamp
        self.database = database

    @staticmethod
    def from_lines(lines: List[str], principal: str, timestamp: str, database: str):
        settings = {}
        raw_hiveql = "".join(lines)
        is_supported = True
        tables = []
        for idx, line in enumerate(lines):
            setting_re_searched = re.search(setting_pattern, line)
            if setting_re_searched:
                key = setting_re_searched.group("key")
                value = setting_re_searched.group("value")
                if key and value:
                    settings[key] = value
            else:
                query = "".join(lines[idx:])
                if re.match("\\s*select", query, re.IGNORECASE):
                    parsed = Parser(query)
                    for table in parsed.tables:
                        if not re.match(db_and_table_pattern, table) and database:
                            table = f"{database}.{table}"
                        tables.append(table)
                break
        return QueryHistory(raw_hiveql, settings, is_supported, tables, principal, timestamp, database)


def get_query_histories(lines: List[str]) -> List[QueryHistory]:
    sessions = {}
    query_session_id = ""
    query = []
    result = []
    for line in lines:
        if query_session_id:
            if re.match(timestamp_pattern, line):
                session = sessions[query_session_id]
                result.append(QueryHistory.from_lines(query, session["principal"],
                                                      session["timestamp"], session["database"]))
                query_session_id = ""
                query = []
            else:
                query.append(line)
        session_id_searched = re.search(session_id_pattern, line)
        if session_id_searched:
            session_id = session_id_searched.groupdict()["session_id"]
            if session_id not in sessions:
                sessions[session_id] = {"database": "", "principal": "pdd-su"}
            principal_searched = re.search(principal_pattern, line)
            if principal_searched:
                sessions[session_id]["principal"] = principal_searched.groupdict()["principal"]
            query_searched = re.search(query_pattern, line)
            if query_searched:
                use_query_searched = re.search(use_database_query_pattern, line)
                if use_query_searched:
                    sessions[session_id]["database"] = use_query_searched.groupdict()["database"]
                    continue
                timestamp_searched = re.search(timestamp_pattern, line)
                if timestamp_searched:
                    sessions[session_id]["timestamp"] = timestamp_searched.groupdict()["timestamp"]
                query_session_id = session_id
                query.append(query_searched.groupdict()["query"])
    return result


if __name__ == "__main__":
    with open("../log/hiveserver2.log.2022-06-13", "r") as logs:
        print(get_query_histories(logs.readlines()))
