from typing import List
from datetime import datetime, timedelta
from .data_util import Data, DataUsage
from .hive_log_parser import *
import re
import ray
import math
from .option_util import *
from .export_util import *
from .hive_util import *
from .log_util import *


LOG_DATE_FORMAT = "%Y-%m-%d"
logger = get_default_logger()


def get_arg_parser():
    parser = get_parser_with_default_args()
    parser.add_argument("--start-date",
                        default=datetime.today() - timedelta(days=8),
                        type=lambda s: datetime.strptime(s, LOG_DATE_FORMAT),
                        help="start date of log file to scrape (yyyy-mm-dd)")
    parser.add_argument("--end-date",
                        type=lambda s: datetime.strptime(s, LOG_DATE_FORMAT),
                        default=datetime.today() - timedelta(days=1),
                        help="end date of log file to scrape (yyyy-mm-dd)",)
    parser.add_argument("--log-dir", type=str, help="directory that contain log files", default="./log")
    parser.add_argument("--database-filter", type=str, help="database name filter to query a hive", default="pdd__db_*")
    parser.add_argument("--log-date-format", type=str, help="date format of log file")
    parser.add_argument("--table-pattern", type=str, help="regex pattern of hive table",
                        default="(?P<data_type>[a-z]+)_(?P<module>[a-z0-9_]+)_(?P<version>v[0-9])")
    parser.add_argument("--database-pattern", type=str, help="regex pattern of hive database",
                        default="pdd__db_(?P<service>[a-z_]+)")
    return parser


def get_data_from_table(table_pattern: str, full_pattern: str, table: str) -> Data:
    if re.match(full_pattern, table):
        return Data.from_pattern(full_pattern, table)
    if re.match(table_pattern, table):
        return Data.from_pattern(table_pattern, table)
    logger.warning(f"table is not matched with pattern: {table}")


def get_hive_databases(
        hive_uri: str, hive_username: str, hive_password: str, database_filter: str) -> List[str]:
    query_results = query_to_hive_with_http_transport(
        hive_uri,
        hive_username,
        hive_password,
        [f"SHOW DATABASES LIKE '{database_filter}'"]
    )
    pdd_databases = []
    for query_result in query_results:
        pdd_databases.append(query_result[0])
    return pdd_databases


def get_hive_tables(
        hive_uri: str, hive_username: str, hive_password: str, database: str) -> List[str]:
    query_results = query_to_hive_with_http_transport(
        hive_uri,
        hive_username,
        hive_password,
        [f"USE {database}", "SHOW TABLES"]
    )
    tables = []
    for query_result in query_results:
        table = query_result[0]
        tables.append(table)
    return tables


def get_hive_tables_group_by_module(
        hive_uri: str, hive_username: str, hive_password: str,
        database_filter: str, pattern: str) -> DataUsage:
    data_usage = DataUsage()
    databases = get_hive_databases(hive_uri, hive_username, hive_password, database_filter)
    for database_name in databases:
        tables = get_hive_tables(hive_uri, hive_username, hive_password, database_name)
        for table_name in tables:
            data = Data.from_pattern(pattern, f"{database_name}.{table_name}")
            data_usage.set_usage_with_data(0, data)
    return data_usage


def get_log_files(start_date: datetime, end_date: datetime, log_dir: str) -> List[str]:
    result = []
    dt = start_date
    while dt <= end_date:
        log_file_name = f"{log_dir}/hiveserver2.log.{dt.strftime(LOG_DATE_FORMAT)}"
        result.append(log_file_name)
        dt += timedelta(days=1)
    return result


@ray.remote
def get_accessed_tables(log_files: List[str], on_read) -> List[QueryHistory]:
    result = []
    for log_file in log_files:
        query_histories = get_query_histories(on_read(log_file))
        result.extend(query_histories)
    return result


def main(args, on_read):
    table_pattern = args.get("table_pattern")
    database_pattern = args.get("database_pattern")
    full_pattern = f"{database_pattern}.{table_pattern}"
    log_files = get_log_files(args.get("start_date"), args.get("end_date"), args.get("log_dir"))
    num_threads = args.get("num_threads")

    n = math.ceil(len(log_files)/num_threads)
    logger.info("getting queries from hiveserver2 log files...")
    task_ids = [get_accessed_tables.remote(log_files[n*i:n*(i+1)], on_read) for i in range(num_threads)]

    data_usage = get_hive_tables_group_by_module(
        args.get("hive_uri"),
        args.get("hive_username"),
        args.get("hive_password"),
        args.get("database_filter"),
        full_pattern
    )
    query_histories = []
    while len(task_ids):
        done, task_ids = ray.wait(task_ids)
        for idx, task_result in enumerate(ray.get(done)):
            logger.info(f"aggregating a hive table usage [{num_threads-len(task_ids)-len(done)+idx+1}/{num_threads}]")
            query_histories.extend(task_result)
            for query_history in task_result:
                for table in query_history.tables:
                    data = get_data_from_table(table_pattern, full_pattern, table)
                    if data:
                        data_usage.set_usage_with_data(data_usage.get_usage_with_data(data)+1, data)
                        total = data_usage.get_usage(data.service, data.version, data.module, "total")
                        data_usage.set_usage(total + 1, data.service, data.version, data.module, "total")
    table_types = ["total", "raw", "compact", "snapshot", "iceberg"]
    logger.info("writing a report...")
    export_data_usage_to_table(data_usage, table_types, ["total"], "hive-log")
    export_unused_items_to_checklist(data_usage, "hive-log")
    export_query_histories_to_table(query_histories, "hive-log")


if __name__ == "__main__":
    def read_file(file: str) -> List[str]:
        with open(file, "r") as f:
            return f.readlines()
    main(vars(get_arg_parser().parse_args()), read_file)
